@@ -32,12 +32,6 @@
 #include <QFile>
 #include <QDir>
 #include <QImage>
-#include <sys/stat.h>
-#ifdef Q_OS_WIN32
-#include <sys/utime.h>
-#else
-#include <utime.h>
-#endif
 #include <cstdio>
 #include <cmath>
 #include <QByteArray>
@@ -92,8 +86,10 @@ static void getPicture(Frame& frame, const FLAC::Metadata::Picture* pic)
  *
  * @param frame frame to get
  * @param pic picture block to set
+ *
+ * @return true if ok.
  */
-static void setPicture(const Frame& frame, FLAC::Metadata::Picture* pic)
+static bool setPicture(const Frame& frame, FLAC::Metadata::Picture* pic)
 {
   Frame::TextEncoding enc;
   PictureFrame::PictureType pictureType = PictureFrame::PT_CoverFront;
@@ -115,7 +111,20 @@ static void setPicture(const Frame& frame, FLAC::Metadata::Picture* pic)
   pic->set_description(
     reinterpret_cast<const FLAC__byte*>(
       static_cast<const char*>(description.toUtf8())));
-  pic->set_data(reinterpret_cast<const FLAC__byte*>(ba.data()), ba.size());
+  const FLAC__byte* data = reinterpret_cast<const FLAC__byte*>(ba.data());
+  int dataLength = ba.size();
+  if (!data || dataLength <= 0) {
+    // Avoid assertion crash in FLAC__metadata_object_picture_set_data().
+    qWarning("FLAC picture data empty");
+    return false;
+  }
+  pic->set_data(data, dataLength);
+  if (pic->get_length() >= (1u << FLAC__STREAM_METADATA_LENGTH_LEN)) {
+    // Avoid assertion crash in FLAC write_metadata_block_header_cb_().
+    qWarning("FLAC picture is too large");
+    return false;
+  }
+  return true;
 }
 #endif // HAVE_FLAC_PICTURE
 
@@ -138,7 +147,7 @@ void FlacFile::readTags(bool force)
   bool priorIsTagInformationRead = isTagInformationRead();
   if (force || !m_fileRead) {
     m_comments.clear();
-    markTag2Unchanged();
+    markTagUnchanged(Frame::Tag_2);
     m_fileRead = true;
     QByteArray fnIn = QFile::encodeName(currentFilePath());
     readFileInfo(m_fileInfo, 0); // just to start invalid
@@ -241,7 +250,7 @@ bool FlacFile::writeTags(bool force, bool* renamed, bool preserve)
     return false;
   }
 
-  if (m_fileRead && (force || isTag2Changed()) && m_chain && m_chain->is_valid()) {
+  if (m_fileRead && (force || isTagChanged(Frame::Tag_2)) && m_chain && m_chain->is_valid()) {
     bool commentsSet = false;
 #ifdef HAVE_FLAC_PICTURE
     bool pictureSet = false;
@@ -249,7 +258,9 @@ bool FlacFile::writeTags(bool force, bool* renamed, bool preserve)
     PictureList::iterator pictureIt = m_pictures.begin();
 #endif
 
-    if (m_chain->status() == FLAC__METADATA_CHAIN_STATUS_NOT_A_FLAC_FILE) {
+    FLAC::Metadata::Chain::Status status = m_chain->status();
+    if (status == FLAC__METADATA_CHAIN_STATUS_NOT_A_FLAC_FILE ||
+        status == FLAC__METADATA_CHAIN_STATUS_ERROR_OPENING_FILE) {
       // This check is done because of a crash in mdit.get_block_type() with an
       // empty file with flac extension. m_chain->status() will set the status
       // to FLAC__METADATA_CHAIN_STATUS_OK (!?), so we have to delete the
@@ -288,8 +299,12 @@ bool FlacFile::writeTags(bool force, bool* renamed, bool preserve)
             FLAC::Metadata::Picture* pic =
               dynamic_cast<FLAC::Metadata::Picture*>(proto);
             if (pic) {
-              setPicture(*pictureIt++, pic);
-              pictureSet = true;
+              if (setPicture(*pictureIt++, pic)) {
+                pictureSet = true;
+              } else {
+                mdit.delete_block(false);
+                pictureRemoved = true;
+              }
             }
             delete proto;
           }
@@ -300,8 +315,7 @@ bool FlacFile::writeTags(bool force, bool* renamed, bool preserve)
       } else if (mdt == FLAC__METADATA_TYPE_PADDING) {
         if (pictureIt != m_pictures.end()) {
           FLAC::Metadata::Picture* pic = new FLAC::Metadata::Picture;
-          setPicture(*pictureIt, pic);
-          if (mdit.set_block(pic)) {
+          if (setPicture(*pictureIt, pic) && mdit.set_block(pic)) {
             ++pictureIt;
             pictureSet = true;
           } else {
@@ -328,8 +342,7 @@ bool FlacFile::writeTags(bool force, bool* renamed, bool preserve)
 #ifdef HAVE_FLAC_PICTURE
         while (pictureIt != m_pictures.end()) {
           FLAC::Metadata::Picture* pic = new FLAC::Metadata::Picture;
-          setPicture(*pictureIt, pic);
-          if (mdit.insert_block_after(pic)) {
+          if (setPicture(*pictureIt, pic) && mdit.insert_block_after(pic)) {
             pictureSet = true;
           } else {
             delete pic;
@@ -343,12 +356,12 @@ bool FlacFile::writeTags(bool force, bool* renamed, bool preserve)
 #ifdef HAVE_FLAC_PICTURE
     if ((commentsSet || pictureSet) &&
         m_chain->write(!pictureRemoved, preserve)) {
-      markTag2Unchanged();
+      markTagUnchanged(Frame::Tag_2);
     }
 #else
     if (commentsSet &&
         m_chain->write(true, preserve)) {
-      markTag2Unchanged();
+      markTagUnchanged(Frame::Tag_2);
     }
 #endif
     else {
@@ -369,118 +382,135 @@ bool FlacFile::writeTags(bool force, bool* renamed, bool preserve)
 
 #ifdef HAVE_FLAC_PICTURE
 /**
- * Check if file has an ID3v2 tag.
+ * Check if file has a tag.
  *
- * @return true if a V2 tag is available.
+ * @param tagNr tag number
+ * @return true if a tag is available.
  * @see isTagInformationRead()
  */
-bool FlacFile::hasTagV2() const
+bool FlacFile::hasTag(Frame::TagNumber tagNr) const
 {
-  return OggFile::hasTagV2() || !m_pictures.empty();
+  return tagNr == Frame::Tag_2 && (OggFile::hasTag(Frame::Tag_2) || !m_pictures.empty());
 }
 
 /**
- * Set a frame in the tags 2.
+ * Set a frame in the tags.
  *
+ * @param tagNr tag number
  * @param frame frame to set
  *
  * @return true if ok.
  */
-bool FlacFile::setFrameV2(const Frame& frame)
+bool FlacFile::setFrame(Frame::TagNumber tagNr, const Frame& frame)
 {
-  if (frame.getType() == Frame::FT_Picture) {
-    int index = frame.getIndex();
-    if (index != -1 && index < static_cast<int>(m_pictures.size())) {
-      PictureList::iterator it = m_pictures.begin() + index;
-      if (it != m_pictures.end()) {
-        Frame newFrame(frame);
-        PictureFrame::setDescription(newFrame, frame.getValue());
-        if (PictureFrame::areFieldsEqual(*it, newFrame)) {
-          it->setValueChanged(false);
-        } else {
-          *it = newFrame;
-          markTag2Changed(Frame::FT_Picture);
+  if (tagNr == Frame::Tag_2) {
+    if (frame.getType() == Frame::FT_Picture) {
+      int index = frame.getIndex();
+      if (index != -1 && index < static_cast<int>(m_pictures.size())) {
+        PictureList::iterator it = m_pictures.begin() + index;
+        if (it != m_pictures.end()) {
+          Frame newFrame(frame);
+          PictureFrame::setDescription(newFrame, frame.getValue());
+          if (PictureFrame::areFieldsEqual(*it, newFrame)) {
+            it->setValueChanged(false);
+          } else {
+            *it = newFrame;
+            markTagChanged(Frame::Tag_2, Frame::FT_Picture);
+          }
+          return true;
         }
-        return true;
       }
     }
   }
-  return OggFile::setFrameV2(frame);
+  return OggFile::setFrame(tagNr, frame);
 }
 
 /**
- * Add a frame in the tags 2.
+ * Add a frame in the tags.
  *
+ * @param tagNr tag number
  * @param frame frame to add
  *
  * @return true if ok.
  */
-bool FlacFile::addFrameV2(Frame& frame)
+bool FlacFile::addFrame(Frame::TagNumber tagNr, Frame& frame)
 {
-  if (frame.getType() == Frame::FT_Picture) {
-    if (frame.getFieldList().empty()) {
-      PictureFrame::setFields(
-        frame, Frame::TE_ISO8859_1, QLatin1String("JPG"), QLatin1String("image/jpeg"),
-        PictureFrame::PT_CoverFront, QLatin1String(""), QByteArray());
+  if (tagNr == Frame::Tag_2) {
+    if (frame.getType() == Frame::FT_Picture) {
+      if (frame.getFieldList().empty()) {
+        PictureFrame::setFields(
+          frame, Frame::TE_ISO8859_1, QLatin1String("JPG"), QLatin1String("image/jpeg"),
+          PictureFrame::PT_CoverFront, QLatin1String(""), QByteArray());
+      }
+      PictureFrame::setDescription(frame, frame.getValue());
+      frame.setIndex(m_pictures.size());
+      m_pictures.push_back(frame);
+      markTagChanged(Frame::Tag_2, Frame::FT_Picture);
+      return true;
     }
-    PictureFrame::setDescription(frame, frame.getValue());
-    frame.setIndex(m_pictures.size());
-    m_pictures.push_back(frame);
-    markTag2Changed(Frame::FT_Picture);
-    return true;
   }
-  return OggFile::addFrameV2(frame);
+  return OggFile::addFrame(tagNr, frame);
 }
 
 /**
- * Delete a frame in the tags 2.
+ * Delete a frame from the tags.
  *
+ * @param tagNr tag number
  * @param frame frame to delete.
  *
  * @return true if ok.
  */
-bool FlacFile::deleteFrameV2(const Frame& frame)
+bool FlacFile::deleteFrame(Frame::TagNumber tagNr, const Frame& frame)
 {
-  if (frame.getType() == Frame::FT_Picture) {
-    int index = frame.getIndex();
-    if (index != -1 && index < static_cast<int>(m_pictures.size())) {
-      m_pictures.removeAt(index);
-      markTag2Changed(Frame::FT_Picture);
-      return true;
+  if (tagNr == Frame::Tag_2) {
+    if (frame.getType() == Frame::FT_Picture) {
+      int index = frame.getIndex();
+      if (index != -1 && index < static_cast<int>(m_pictures.size())) {
+        m_pictures.removeAt(index);
+        markTagChanged(Frame::Tag_2, Frame::FT_Picture);
+        return true;
+      }
     }
   }
-  return OggFile::deleteFrameV2(frame);
+  return OggFile::deleteFrame(tagNr, frame);
 }
 
 /**
- * Remove ID3v2 frames.
+ * Remove frames.
  *
+ * @param tagNr tag number
  * @param flt filter specifying which frames to remove
  */
-void FlacFile::deleteFramesV2(const FrameFilter& flt)
+void FlacFile::deleteFrames(Frame::TagNumber tagNr, const FrameFilter& flt)
 {
+  if (tagNr != Frame::Tag_2)
+    return;
+
   if (flt.areAllEnabled() || flt.isEnabled(Frame::FT_Picture)) {
     m_pictures.clear();
-    markTag2Changed(Frame::FT_Picture);
+    markTagChanged(Frame::Tag_2, Frame::FT_Picture);
   }
-  OggFile::deleteFramesV2(flt);
+  OggFile::deleteFrames(tagNr, flt);
 }
 
 /**
- * Get all frames in tag 2.
+ * Get all frames in tag.
  *
+ * @param tagNr tag number
  * @param frames frame collection to set.
  */
-void FlacFile::getAllFramesV2(FrameCollection& frames)
+void FlacFile::getAllFrames(Frame::TagNumber tagNr, FrameCollection& frames)
 {
-  OggFile::getAllFramesV2(frames);
-  int i = 0;
-  for (PictureList::iterator it = m_pictures.begin();
-       it != m_pictures.end();
-       ++it) {
-    (*it).setIndex(i++);
-    updateMarkedState(*it);
-    frames.insert(*it);
+  OggFile::getAllFrames(tagNr, frames);
+  if (tagNr == Frame::Tag_2) {
+    int i = 0;
+    for (PictureList::iterator it = m_pictures.begin();
+         it != m_pictures.end();
+         ++it) {
+      (*it).setIndex(i++);
+      updateMarkedState(*it);
+      frames.insert(*it);
+    }
   }
 }
 #endif // HAVE_FLAC_PICTURE
